# NEST Firmware - Arduino Version (Wearable Prototype)

This repository contains the firmware for the **NEST (Non-intrusive dEviceS for Telemedicine)** project, developed for a **wearable prototype** on the **Arduino** platform. The firmware controls the monitoring of **heart rate (BPM)** through the **MAX30102** sensor and the **nRF52840** microcontroller with integrated **Bluetooth Low Energy (BLE)**. It also includes a touch interface to interact with the device’s functionalities, such as rescue confirmation, monitoring, and sending data via BLE.

## Repository Structure

- **/bootloader**: Bootloader files for the nRF52840, which need to be manually flashed.
- **/firmware**: Source code for the firmware developed in Arduino.
- **/libraries**: Required libraries such as **Ucglib.h**, **algorithm_by_RF.h**, and **max30102.h** to control the TFT screen and heart rate sensor.

## Firmware Description

This firmware is designed for a medical wearable that allows heart rate monitoring and user interaction through a touchscreen. The main functionalities of the firmware include:

1. **BPM Monitoring**: Uses the **MAX30102** sensor to measure heart rate (BPM). The data is processed and displayed on the TFT screen.
2. **TFT Screen and Touch Control**: The user interface includes a main menu with options to monitor vital signs, interact with a pain diary, and confirm rescue data sending. All this is controlled through a TFT screen managed by the **Ucglib.h** library and a touch system based on the **TouchScreen.h** library.
3. **Bluetooth Connectivity**: Implements BLE using **Bluefruit.h** to send notifications and monitoring data to connected devices. It includes custom BLE characteristics for sending alerts.
4. **Screen Backlight Control**: The screen backlight pin is controlled with PWM, allowing brightness adjustment based on the device's state.
5. **Dynamic Heart Rate Update**: The BPM value is updated in real time and displayed on the screen. It can also be sent via BLE if a device is connected.

## Main Functions

- **Main Screen**: Presents three options: **Pain Diary**, **Rescue**, and **Monitoring**.
- **Rescue Confirmation**: Allows the user to confirm or cancel the rescue data transmission.
- **Data Sending**: Shows a sending animation and notifies when data has been successfully sent via BLE.
- **Heart Rate Monitoring**: Displays the current BPM value on the screen and allows returning to the main menu by touching the screen.

### Libraries Included in the Project

The project includes the following custom libraries found in the `/libraries` directory:

- **Ucglib.h**: Used to control the TFT screen and implement graphical interfaces.
- **algorithm_by_RF.h**: Processes data from the **MAX30102** sensor to calculate heart rate.
- **max30102.h**: Handles the operations of the **MAX30102** sensor for reading biometric data.

Other required libraries, such as **Adafruit_TinyUSB.h**, **TouchScreen.h**, **Bluefruit.h**, and **SPI.h**, can be installed directly from the **Arduino IDE** library manager.

## Arduino Programming Requirements

### Install Arduino Core for nRF52840

1. **Download and install Arduino IDE** (at least v1.6.12): [Arduino IDE](https://www.arduino.cc/en/software).
2. Open the Arduino IDE and go to **Preferences**.
3. In **Additional Board Manager URLs**, add:  
   `https://adafruit.github.io/arduino-board-index/package_adafruit_index.json`
4. Restart the Arduino IDE.
5. Go to **Tools > Board > Board Manager**, search for "Adafruit nRF52" and install **Adafruit nRF52 by Adafruit**.
6. Once installed, select **Adafruit Feather nRF52840 Express** from **Tools > Board** to configure the system.

## Bootloader

The bootloader must be manually flashed. In this repository, you will find the file that needs to be flashed to the nRF52840 in the `/bootloader` directory. Follow the instructions below to flash the bootloader.

### Flash a New Bootloader

To flash the bootloader from the Arduino IDE, you will need the following tools installed and available in your system path:

- [Segger JLink Software and Documentation Pack](https://www.segger.com/downloads/jlink)
- [Nordic nRF5x Command Line Tools](https://www.nordicsemi.com/Software-and-Tools/Development-Tools/nRF-Command-Line-Tools)

Ensure you can run `nrfjprog` from your terminal or command prompt.

### Manually Flash the Bootloader with nrfjprog

The bootloader hex file can be found in the `/bootloader` folder. Run the following commands from the folder where the bootloader is located:

```bash
$ nrfjprog -e -f nrf52
$ nrfjprog --program feather_nrf52840_bootloader.hex -f nrf52
$ nrfjprog --reset -f nrf52
```

### Arduino Configurations

Once the bootloader is flashed, follow these configurations in the Arduino IDE:

1. Select **Tools > Board > Adafruit Feather nRF52840 Express**.
2. Select the correct COM port for the board.
3. In **Tools > Programmer**, choose **J-Link for Bluefruit nRF52**.
4. Then, select **Tools > Burn Bootloader** to finalize the setup process.


