bool checkTouch() {
  TSPoint p = ts.getPoint();
  pinMode(XM, OUTPUT);
  pinMode(YP, OUTPUT);
  bool screenTouched = false;

  // Check if screen was touched by pressure threshold
  if (p.z > ts.pressureThreshhold) {
    int x = map(p.x, TS_MINX, TS_MAXX, 0, 128);
    int y = map(p.y, TS_MINY, TS_MAXY, 0, 160);
    screenTouched = true;

    // Switch case for handling different screens based on the current screen state
    switch (currentScreen) {
      case MAIN_MENU:
        if (y > 30 && y < 66) {  
          drawPainDiaryScreen();  // Open Pain Diary screen
        } else if (y > 75 && y < 111) {  
          drawRescueConfirmationScreen();
        } else if (y > 120 && y < 156) {  
          drawHeartRateScreen();
        }
        break;

      case PAIN_DIARY:
        // Select pain intensity level and proceed accordingly
        if (y > 40 && y < 70) { 
          Serial.println("Ausencia de dolor seleccionada");
          selectedPainLevel = 0;  // No pain
          drawInterface();
        } else if (y > 80 && y < 110) { 
          Serial.println("Dolor leve seleccionado");
          selectedPainLevel = 1;  // Mild pain
          drawPainDescriptionScreen();  // Proceed to pain description screen
        } else if (y > 120 && y < 150) {  
          Serial.println("Dolor moderado seleccionado");
          selectedPainLevel = 2;  // Moderate pain
          drawPainDescriptionScreen();
        } else if (y > 160 && y < 190) {  
          Serial.println("Dolor severo seleccionado");
          selectedPainLevel = 3;  // Severe pain
          drawPainDescriptionScreen();
        }
        break;

      case PAIN_DESCRIPTION:
        // Select type of pain description and proceed to daily limitation screen
        if (y > 40 && y < 70) { 
          painDescription = 0;  // Changing intensity
          drawDailyLimitationScreen();
        } else if (y > 80 && y < 110) { 
          painDescription = 1;  // Constant intensity
          drawDailyLimitationScreen();
        } else if (y > 120 && y < 150) {  
          painDescription = 2;  // Stabbing pain
          drawDailyLimitationScreen();
        } else if (y > 160 && y < 190) {  
          painDescription = 3;  // Pain when standing
          drawDailyLimitationScreen();
        }
        break;

      case RESCUE_CONFIRM:
        // Confirm or cancel rescue operation
        if (x > 20 && x < 60 && y > 80 && y < 120) { 
          drawInterface();  
        } else if (x > 70 && x < 110 && y > 80 && y < 120) {  
          drawSendingDataScreen();  
        }
        break;

      case SENDING_DATA:
        drawInterface();  // Return to main interface after sending data
        break;

      case SURVEY_SENT:
        if (x > 80 && x < 120 && y > 80 && y < 100)
          drawInterface();
        break;

      case HEART_RATE:
        if (x > 20 && x < 60 && y > 80 && y < 120) {  
          drawInterface();
        }
        break;

      case DAILY_LIMITATION:
        // Select daily limitation level and start HR measurement screen
        if (y > 40 && y < 70) { 
          dailyLimitation = 0;  // No limitation
        } else if (y > 80 && y < 110) { 
          dailyLimitation = 1;  // Slightly limiting
        } else if (y > 120 && y < 150) {  
          dailyLimitation = 2;  // Moderate limitation
        } else if (y > 160 && y < 190) {  
          dailyLimitation = 3;  // Total limitation
        }
        drawHRMeasurementScreen();  // Proceed to "Measuring HR" screen
        break;

      case HR_MEASUREMENT:
        if (x > 20 && x < 60 && y > 80 && y < 120)  {  // Coordinates for purple button
          goToPreviousScreen();  // Go back to the previous screen
        }
        break;

      case HR_ERROR:
        if (x > 20 && x < 60 && y > 80 && y < 120) {  // Coordinates for purple button in error screen
          drawHRMeasurementScreen();  // Retry HR measurement screen
        }
        break;

      case HR_RESULT:
        if (x > 20 && x < 60 && y > 80 && y < 120) {  // Left button
          drawHRMeasurementScreen();  // Return to HR measurement screen
        } else if (x > 80 && x < 120 && y > 80 && y < 120) {  // Right button
          drawDataConfirmationScreen();  // Proceed to data confirmation screen
        }
        break;

      case DATA_CONFIRMATION:
        // Confirm or cancel data sending
        if (x > 20 && x < 60 && y > 80 && y < 120) { 
          drawInterface();  
        } else if (x > 70 && x < 110 && y > 80 && y < 120) {  
          drawDataSendingScreen();  
        }
        break;

      default:
        break;
    }
  }
  return screenTouched;
}


void sendSurveyDataViaBLE() {
  // Compile survey data and BPM for BLE transmission
  String dataToSend = "BPM: " + String(currentBpm) + 
                      ", Intensidad: " + String(selectedPainLevel) + 
                      ", Descripción: " + String(painDescription) + 
                      ", Limitación: " + String(dailyLimitation);

  const char* alertMsg = dataToSend.c_str();  // Convert data to BLE-compatible format

  if (Bluefruit.connected()) {
    // Attempt to send BLE notification with data
    if (alertCharacteristic.notify(alertMsg, strlen(alertMsg))) {
      Serial.println("Datos de encuesta enviados vía BLE");
      drawSurveySentScreen();  // Show "Survey Sent" screen after sending
    } else {
      Serial.println("Error al enviar los datos de la encuesta");
      drawSurveySentScreen();  // Show "Survey Sent" screen after sending
    }
  } else {
    Serial.println("No hay dispositivos conectados.");
    drawSurveySentScreen();  // Show "Survey Sent" screen after sending
  }
}


void drawButton(int x, int y, int w, int h, const char* text, uint16_t color) {
  ucg.setColor(255, 255, 255);  // Draw white border for button
  ucg.drawRBox(x, y, w, h, 5);

  // Convert RGB565 color to RGB888
  uint8_t r = (color >> 11) & 0x1F;
  uint8_t g = (color >> 5) & 0x3F;
  uint8_t b = color & 0x1F;

  // Scale color values to 0-255
  r = (r * 255) / 31;
  g = (g * 255) / 63;
  b = (b * 255) / 31;

  // Set and draw button color with rounded corners
  ucg.setColor(r, g, b);
  ucg.drawRBox(x + 2, y + 2, w - 4, h - 4, 5);

  // Draw button text in white
  ucg.setColor(255, 255, 255);
  ucg.setFont(ucg_font_6x10_tr);
  int textX = x + (w - ucg.getStrWidth(text)) / 2;
  int textY = y + h / 2 + 3;
  ucg.setPrintPos(textX, textY);
  ucg.print(text);
}


void updateHeartRate() {
  static unsigned long lastUpdateTime = 0;  
  int32_t n_heart_rate; 
  int8_t ch_hr_valid;   
  float n_spo2, ratio, correl;  
  int8_t ch_spo2_valid;  
  int32_t i;

  // Read data from MAX30102 sensor
  for (i = 0; i < BUFFER_SIZE; i++) {
    while (digitalRead(OXI_INT) == 1);  
    maxim_max30102_read_fifo((aun_red_buffer + i), (aun_ir_buffer + i));  
  }

  // Process heart rate and oxygen saturation data
  rf_heart_rate_and_oxygen_saturation(aun_ir_buffer, BUFFER_SIZE, aun_red_buffer, &n_spo2, &ch_spo2_valid, &n_heart_rate, &ch_hr_valid, &ratio, &correl);

  if (ch_hr_valid) {
    currentBpm = n_heart_rate;
  } else {
    currentBpm = -1;  
  }

  // Update display if on the HEART_RATE screen
  if (currentScreen == HEART_RATE) {
    ucg.setColor(0, 0, 0);  // Black to clear previous reading
    ucg.drawBox(40, 50, 80, 35);  

    ucg.setColor(255, 255, 255);  // White for new BPM reading
    ucg.setFont(ucg_font_ncenR24_tr);
    ucg.setPrintPos(60, 80);  

    if (currentBpm >= 0) {
      ucg.print(currentBpm);  // Display current BPM
    } else {
      ucg.print("  -");  // Display dash if BPM is invalid
    }
  }
}




