#include <Adafruit_TinyUSB.h>
#include <SPI.h>
#include "Ucglib.h"

// Define los pines según tu configuración
#define TFT_CS     33  // Pin de selección de chip
#define TFT_RST    2   // Pin de reset
#define TFT_DC     25  // Pin de comando/datos
#define TFT_MOSI   24  // Pin de datos de SPI
#define TFT_SCLK   26  // Pin de reloj de SPI
#define LIGHT_PIN  10  // Pin para el control de la luz de fondo

// Configuración de la pantalla Ucglib
Ucglib_ST7735_18x128x160_SWSPI ucg(TFT_SCLK, TFT_MOSI, TFT_DC, TFT_CS, TFT_RST);

void setup(void) {
  pinMode(LIGHT_PIN, OUTPUT);    // Configura el pin de la luz de fondo como salida
  analogWrite(LIGHT_PIN, 255);   // Establece la luz de fondo al máximo

  ucg.begin(UCG_FONT_MODE_TRANSPARENT);  // Inicia la pantalla
  ucg.clearScreen();                     // Limpia la pantalla

  // Configurar la frecuencia del SPI (Ejemplo: 8 MHz)
  SPI.setClockDivider(SPI_CLOCK_DIV2); // Ajusta este valor según tus necesidades y capacidades del hardware

  drawButton(5, 20, 118, 40, "Diario del dolor", 0x2104);  // Gris oscuro
  drawButton(5, 70, 118, 40, "Rescate", 0x2104);           // Gris oscuro
  drawButton(5, 120, 118, 40, "Monitoreo", 0x2104);        // Gris oscuro
}

void loop(void) {
  // Aquí puedes agregar cualquier código que necesites ejecutar repetidamente
}

void drawButton(int x, int y, int w, int h, const char* text, uint16_t color) {
  // Dibuja el fondo del botón
  ucg.setColor(255, 255, 255);  // Blanco para el borde
  ucg.drawRBox(x, y, w, h, 5);  // Caja redondeada
  ucg.setColor(color >> 8, color & 0xFF, color & 0xFF);  // Color de fondo del botón
  ucg.drawRBox(x + 2, y + 2, w - 4, h - 4, 5);  // Caja redondeada interior

  // Dibuja el texto
  ucg.setColor(255, 255, 255);  // Blanco para el texto
  ucg.setFont(ucg_font_6x10_tr);  // Fuente para el texto
  ucg.setPrintPos(x + 10, y + h / 2 + 4);  // Centra el texto verticalmente y margen izquierdo
  ucg.print(text);
}

